#include <stdio.h>

#include "hash.h"
#include "tunes.h"

void
error (const char *s)
{
  fprintf (stderr, "%s\n", s);
  exit (1);
}

static objptr_t freeptr = 0;

objptr_t
allocate (void)
{
  objptr_t obj;

  obj = freeptr++;
  if ( obj >= CORESIZE )
    error ("Need to GC!");
  return obj;
}

objptr_t
a_unit (void)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_UNIT;
  return obj;
}

objptr_t
lookup (hashv_t name, objptr_t env)
     /* Lookup a name in an environment. */
{
  if ( core[env].type == O_ENVNULL )
    error ("Unbound variable.");
  else if ( core[env].type != O_ENV )
    error ("INTERNAL ERROR: Environment expected!");
  else if ( core[core[env].d.env_v.def].type != O_DEF )
    error ("INTERNAL ERROR: Frame expected!");
  else if ( core[core[env].d.env_v.def].d.def_v.name == name )
    return core[core[env].d.env_v.def].d.def_v.val;
  else
    return lookup (name, core[env].d.env_v.seq);
}

objptr_t
eval (objptr_t expr, objptr_t env, objptr_t cont)
     /* Evaluate expr in environment env.  Either pass the result to
      * cont (via invoke()) or return a task that must be run (via
      * step()) to continue the computation. */
{
  if ( core[expr].type == O_VAR )
    return invoke (cont, lookup (core[expr].d.var_v, env));
  else if ( core[expr].type == O_APPLY )
    {
      objptr_t task;

      task = allocate ();
      core[task].type = O_TASK_APPLY_0;
      core[task].d.task_apply_0_v.rator = core[expr].d.apply_v.rator;
      core[task].d.task_apply_0_v.rand = core[expr].d.apply_v.rand;
      core[task].d.task_apply_0_v.env = env;
      core[task].d.task_apply_0_v.cont = cont;
      return task;
    }
  else if ( core[expr].type == O_PROGN )
    {
      objptr_t task;

      task = allocate ();
      core[task].type = O_TASK_PROGN_0;
      core[task].d.task_progn_0_v.primus = core[expr].d.progn_v.primus;
      core[task].d.task_progn_0_v.pares = core[expr].d.progn_v.pares;
      core[task].d.task_progn_0_v.env = env;
      core[task].d.task_progn_0_v.cont = cont;
      return task;
    }
  else if ( core[expr].type == O_QUOTE )
    return invoke (cont, core[expr].d.quote_v);
  else if ( core[expr].type == O_LAMBDA )
    {
      objptr_t fun;

      fun = allocate ();
      core[fun].type = O_USERFUNC;
      core[fun].d.userfunc_v.param = core[expr].d.lambda_v.param;
      core[fun].d.userfunc_v.env = env;
      core[fun].d.userfunc_v.body = core[expr].d.lambda_v.body;
      return invoke (cont, fun);
    }
  else
    /* Deprecated: please use QUOTE instead. */
    return invoke (cont, expr);
}

objptr_t
step (objptr_t task)
     /* Advance a task by one step */
{
  if ( core[task].type == O_TASK_EVAL )
    return eval (core[task].d.task_eval_v.expr,
		 core[task].d.task_eval_v.env,
		 core[task].d.task_eval_v.cont);
  else if ( core[task].type == O_TASK_APPLY )
    return apply (core[task].d.task_apply_v.rator,
		  core[task].d.task_apply_v.rand,
		  core[task].d.task_apply_v.cont);
  else if ( core[task].type == O_TASK_PROGN_0 )
    {
      objptr_t newcont;

      newcont = allocate ();
      core[newcont].type = O_CONT_EVAL;
      core[newcont].d.cont_eval_v.expr
	= core[task].d.task_progn_0_v.pares;
      core[newcont].d.cont_eval_v.env
	= core[task].d.task_progn_0_v.env;
      core[newcont].d.cont_eval_v.cont
	= core[task].d.task_progn_0_v.cont;
      return eval (core[task].d.task_progn_0_v.primus,
		   core[task].d.task_progn_0_v.env,
		   newcont);
    }
  else if ( core[task].type == O_TASK_APPLY_0 )
    {
      objptr_t newcont;

      newcont = allocate ();
      core[newcont].type = O_CONT_APPLY_1;
      core[newcont].d.cont_apply_1_v.rand
	= core[task].d.task_apply_0_v.rand;
      core[newcont].d.cont_apply_1_v.env
	= core[task].d.task_apply_0_v.env;
      core[newcont].d.cont_apply_1_v.cont
	= core[task].d.task_apply_0_v.cont;
      return eval (core[task].d.task_apply_0_v.rator,
		   core[task].d.task_apply_0_v.env,
		   newcont);
    }
  else if ( core[task].type == O_TASK_APPLY_1 )
    {
      objptr_t newcont;

      newcont = allocate ();
      core[newcont].type = O_CONT_APPLY;
      core[newcont].d.cont_apply_v.rator
	= core[task].d.task_apply_1_v.rator;
      core[newcont].d.cont_apply_v.cont
	= core[task].d.task_apply_1_v.cont;
      return eval (core[task].d.task_apply_1_v.rand,
		   core[task].d.task_apply_1_v.env,
		   newcont);
    }
  else
    error ("INTERNAL ERROR: Trying to run a non-task.");
}

objptr_t
invoke (objptr_t cont, objptr_t val)
     /* Invoke a continuation: return a task. */
{
  if ( core[cont].type == O_CONT_FINISH )
    {
      fprintf (stderr, "Execution terminated.\n");
      exit (0);
    }
  else if ( core[cont].type == O_CONT_EVAL )
    {
      objptr_t task;

      task = allocate ();
      core[task].type = O_TASK_EVAL;
      core[task].d.task_eval_v.expr = core[cont].d.cont_eval_v.expr;
      core[task].d.task_eval_v.env = core[cont].d.cont_eval_v.env;
      core[task].d.task_eval_v.cont = core[cont].d.cont_eval_v.cont;
      return task;
    }
  else if ( core[cont].type == O_CONT_APPLY )
    {
      objptr_t task;

      task = allocate ();
      core[task].type = O_TASK_APPLY;
      core[task].d.task_apply_v.rator = core[cont].d.cont_apply_v.rator;
      core[task].d.task_apply_v.rand = val;
      core[task].d.task_apply_v.cont = core[cont].d.cont_apply_v.cont;
      return task;
    }
  else if ( core[cont].type == O_CONT_APPLY_1 )
    {
      objptr_t task;

      task = allocate ();
      core[task].type = O_TASK_APPLY_1;
      core[task].d.task_apply_1_v.rator = val;
      core[task].d.task_apply_1_v.rand
	= core[cont].d.cont_apply_1_v.rand;
      core[task].d.task_apply_1_v.env
	= core[cont].d.cont_apply_1_v.env;
      core[task].d.task_apply_1_v.cont
	= core[cont].d.cont_apply_1_v.cont;
      return task;
    }
  else
    error ("INTERNAL ERROR: Tried to invoke a non-continuation.");
}

objptr_t
apply_pf_print (objptr_t rand, objptr_t cont)
{
  if ( core[rand].type == O_INT )
    printf ("%d\n", core[rand].d.int_v);
  else if ( core[rand].type == O_CHAR )
    printf ("%c", core[rand].d.char_v);
  else
    error ("Wrong type argument to primitive-function-print.");
  return invoke (cont, a_unit ());
}

objptr_t
apply_pf_i (objptr_t rand, objptr_t cont)
{
  return invoke (cont, rand);
}

objptr_t
apply_pf_k (objptr_t rand, objptr_t cont)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_PRIMFUNC1;
  core[obj].d.primfunc1_v.nm = PF1_K;
  core[obj].d.primfunc1_v.data1 = rand;
  return invoke (cont, obj);
}

objptr_t
apply_pf_s (objptr_t rand, objptr_t cont)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_PRIMFUNC1;
  core[obj].d.primfunc1_v.nm = PF1_S;
  core[obj].d.primfunc1_v.data1 = rand;
  return invoke (cont, obj);
}

objptr_t
apply_pf_call_cc (objptr_t rand, objptr_t cont)
{
  objptr_t task;

  task = allocate ();
  core[task].type = O_TASK_APPLY;
  core[task].d.task_apply_v.rator = rand;
  core[task].d.task_apply_v.rand = cont;
  core[task].d.task_apply_v.cont = cont;
  return task;
}

objptr_t
apply_primfunc (primfunc_t fn, objptr_t rand, objptr_t cont)
{
  switch ( fn )
    {
    case PF_CALL_CC: return apply_pf_call_cc (rand, cont);
    case PF_PRINT: return apply_pf_print (rand, cont);
    case PF_I: return apply_pf_i (rand, cont);
    case PF_K: return apply_pf_k (rand, cont);
    case PF_S: return apply_pf_s (rand, cont);
    default: error ("INTERNAL ERROR: Unknown primitive function type.");
    }
}

objptr_t
apply_pf1_k (objptr_t data1, objptr_t rand, objptr_t cont)
{
  return data1; /* rand is gc-fodder */
}

objptr_t
apply_pf1_s (objptr_t data1, objptr_t rand, objptr_t cont)
{
  objptr_t obj;

  obj = allocate ();
  core[obj].type = O_PRIMFUNC2;
  core[obj].d.primfunc2_v.nm = PF2_S;
  core[obj].d.primfunc2_v.data1 = data1;
  core[obj].d.primfunc2_v.data2 = rand;
  return invoke (cont, obj);
}

objptr_t
apply_primfunc1 (primfunc_t fn, objptr_t data1, objptr_t rand, objptr_t cont)
{
  switch ( fn )
    {
    case PF1_K: return apply_pf1_k (data1, rand, cont);
    case PF1_S: return apply_pf1_s (data1, rand, cont);
    default: error ("INTERNAL ERROR: Unknown primitive function (1) type.");
    }
}

objptr_t
apply_pf2_s (objptr_t data1, objptr_t data2,
	     objptr_t rand, objptr_t cont)
{
  /* XXX - FIXME: THIS IS COMPLETELY WRONG */
  objptr_t obj1, obj2;

  obj1 = apply (data1, rand, cont);
  obj2 = apply (data2, rand, cont);
  return apply (obj1, obj2, cont);
}

objptr_t
apply_primfunc2 (primfunc_t fn, objptr_t data1, objptr_t data2,
		 objptr_t rand, objptr_t cont)
{
  switch ( fn )
    {
    case PF2_S: return apply_pf2_s (data1, data2, rand, cont);
    default: error ("INTERNAL ERROR: Unknown primitive function (1) type.");
    }
}

objptr_t
extend_env (hashv_t name, objptr_t val, objptr_t env)
{
  objptr_t frame, newenv;

  frame = allocate ();
  core[frame].type = O_DEF;
  core[frame].d.def_v.name = name;
  core[frame].d.def_v.val = val;
  newenv = allocate ();
  core[newenv].type = O_ENV;
  core[newenv].d.env_v.def = frame;
  core[newenv].d.env_v.seq = env;
  return newenv;
}

objptr_t
apply (objptr_t rator, objptr_t rand, objptr_t cont)
{
  if ( core[rator].type == O_PRIMFUNC )
    return apply_primfunc (core[rator].d.primfunc_v.nm,
			   rand, cont);
  else if ( core[rator].type == O_PRIMFUNC1 )
    return apply_primfunc1 (core[rator].d.primfunc1_v.nm,
			    core[rator].d.primfunc1_v.data1,
			    rand, cont);
  else if ( core[rator].type == O_PRIMFUNC2 )
    return apply_primfunc2 (core[rator].d.primfunc2_v.nm,
			    core[rator].d.primfunc2_v.data1,
			    core[rator].d.primfunc2_v.data2,
			    rand, cont);
  else if ( core[rator].type == O_USERFUNC )
    return eval (core[rator].d.userfunc_v.body,
		 extend_env (core[rator].d.userfunc_v.param, rand,
			     core[rator].d.userfunc_v.env),
		 cont);
  else if ( core[rator].type >= O_CONT_FINISH
	    && core[rator].type < O_TASK_EVAL )
    return invoke (rator, rand); /* Calling continuation: discard cont */
  else
    error ("Wrong type to apply.");
}

int
main (void)
{
  objptr_t final;
  objptr_t task;

  yyparse ();
  create_globenv ();
  final = allocate ();
  core[final].type = O_CONT_FINISH;
  task = eval (mainprog, globenv, final);
  while ( 1 )
    task = step (task);
  return 0;
}
